use actix_web::{
    error::{ErrorBadRequest, ErrorInternalServerError}, 
    web::Data, HttpRequest, dev::Payload,
    FromRequest, Error
};

use jsonwebtoken::{Header, EncodingKey, DecodingKey, Validation};
use serde::{Serialize, Deserialize};
use std::future::{self, ready};
use chrono::Utc;

use crate::State;

static WEEK: i64 = 60 * 60 * 24 * 7;

#[derive(Debug, Serialize, Deserialize)]
pub struct Claims {
    pub user_id: i32,
    pub exp: usize,
}

impl Claims {
    pub fn generate_token(secret: &str, user_id: i32) -> String {
        let now = Utc::now().timestamp();

        let payload = Claims {
            exp: (now + WEEK) as usize,
            user_id,
        };
        
        jsonwebtoken::encode(
            &Header::default(), 
            &payload, 
            &EncodingKey::from_secret(secret.as_ref())
        ).unwrap()
    }
}

#[derive(Deserialize, Debug, Serialize)]
pub struct AuthorizedUser {
    pub token: String,
    pub user_id: i32,
}

impl FromRequest for AuthorizedUser {
    type Error = Error;
    type Future = std::future::Ready<Result<Self, Self::Error>>;

    fn from_request(req: &HttpRequest, _: &mut Payload) -> Self::Future {
        let state = req.app_data::<Data<State>>();
        let header = req.headers().get("Authorization");
        if header.is_none() {
            return future::ready(Err(ErrorBadRequest("no Authorization header specified")))
        }

        if state.is_none() {
            return future::ready(Err(ErrorInternalServerError("cannot retrieve JWT token")))
        }

        let state = state.unwrap();
        match header.unwrap().to_str() {
            Ok(v) => {
                let parts: Vec<&str> = v.split(" ").collect();

                if parts.len() != 2 {
                    return ready(Err(ErrorBadRequest("unauthorized")));
                }

                let claims = jsonwebtoken::decode::<Claims>(
                    parts[1], 
                    &DecodingKey::from_secret(state.config.jwt_secret.as_ref()), 
                    &Validation::default()
                );

                if claims.is_err() {
                    return ready(Err(ErrorBadRequest("unauthorized")));
                }

                let claims = claims.unwrap().claims;
                let authorized_user = AuthorizedUser {
                    user_id: claims.user_id,
                    token: v.to_string()
                };

                ready(Ok(authorized_user))
            }
            Err(e) => ready(Err(ErrorBadRequest(e))),
        }
    }
}